#include "Info.h"

FunctionInfo::FunctionInfo(FuncDef *expr, ClassInfoPtr klass)
: VarInfo(expr->pident_, nullptr), block(expr->block_), klass(klass)
{
	ListType *funArgs = new ListType();
	type = make_shared<Fun>(expr->type_, funArgs);
	arguments.reserve(expr->listarg_->size());
	for (auto arg : *expr->listarg_) {
		VarInfoPtr var = make_shared<VarInfo>((Ar*)arg);
		var->ident->var = var;
		arg->pident_->var = var;
		arguments.emplace_back(var);
		funArgs->emplace_back(arg->type_);
	}
}

Binding::Binding(shared_ptr<Binding> parent)
	: variables(parent ? &parent->variables : nullptr),
	classes(parent ? &parent->classes : nullptr, &variables),
	functions(parent ? &parent->functions : nullptr, &variables),
	parent(parent)
{

}

ClassInfo::ClassInfo(PIdent *ident, BindingPtr parent /*= nullptr*/)
	: VarInfo(ident),
	Binding(parent),
	size(0), functionCount(0)
{
	type = make_shared<ClassT>(ident->clone());
	type->binding = parent;
}

size_t ClassInfo::calculateSize() {
	if (size) return size;
	size = 4; // pointer to virtual methods table

	if (auto p = getClassParent()) { // parent already has virt pointer included
		p->calculateSize();
		size = p->size;
		functionCount = p->functionCount;
	}

	for (const auto& var : variables) {
		var->offset = size;
		size += var->type->size();
	}

	for (const auto& f : functions) {
		f->offset = functionCount++;
	}

	return size;
};

Scope::Scope()
	: currentClass()
{
}

