#ifndef PARSEERROR_HEADER
#define PARSEERROR_HEADER
#include "TypeDefs.h"
#include <string>
#include <stdexcept>
#include <memory>
#include <vector>
using namespace std;

// class VarInfo;
// using VarInfoPtr = shared_ptr<VarInfo>;

class ParseError : std::runtime_error {
protected:
	string msg;
	ParseError() : runtime_error("ParseError") {}
public:
	int line;
	ParseError(string reason, int line = -1);
	ParseError(string reason, Visitable *expr);
	ParseError(string reason, VarInfo &var);

	ParseError(const ParseError &e, Visitable *param) : ParseError(e.what(), param) { if (e.line != -1) line = e.line; };
	ParseError(const ParseError &e, VarInfo &param) : ParseError(e.what(), param) { if (e.line != -1) line = e.line; };

	virtual const char * what() const _GLIBCXX_TXN_SAFE_DYN _GLIBCXX_NOTHROW { return msg.data(); }
};

class PIdent;

class RedefinedError : public ParseError {
public:
	RedefinedError(PIdent *ident, VarInfoPtr orig);
};

class UndefinedError : public ParseError {
public:
	UndefinedError(PIdent *ident);
};

class InvalidTypeError : public ParseError {
public:
	InvalidTypeError(Type &expected, vector<shared_ptr<Type>> received, Visitable *expr = NULL);
	InvalidTypeError(Type &expected, vector<Type*> received, Visitable *expr = NULL);
};

#endif
