#ifndef ZAD2_UTIL_H
#define ZAD2_UTIL_H
#include <set>
using namespace std;

template<typename T>
set<T> operator+ (const set<T> &l, const set<T> &r) {
	set<T> ret;
	set_union(l.begin(), l.end(), r.begin(), r.end(), std::inserter(ret, ret.begin()));
	return ret;
}

template<typename T>
set<T> operator- (const set<T> &l, const set<T> &r) {
	set<T> ret;
	set_difference(l.begin(), l.end(), r.begin(), r.end(), std::inserter(ret, ret.begin()));
	return ret;
}

template<typename T>
set<T> operator^(const set<T> &l, const set<T> &r) {
	set<T> ret;
	set_symmetric_difference(l.begin(), l.end(), r.begin(), r.end(), std::inserter(ret, ret.begin()));
	return ret;
}

template<typename T>
set<T>& operator+=(set<T> &l, const set<T> &r) {
	auto diff = r - l;
	l.insert(diff.begin(), diff.end());
	return l;
}

template<typename T>
set<T>& operator-=(set<T> &l, const set<T> &r) {
	auto diff = l - r;
	return l = diff;
}

template<typename T>
set<T>& operator^=(set<T> &l, const set<T> &r) {
	auto diff = l ^ r;
	return l = diff;
}

template<typename T>
bool operator==(const set<T> &l, const set<T> &r) {
	return (r^l).empty();
}

template<typename T>
bool operator!=(const set<T> &l, const set<T> &r) {
	return !(r == l);
}

template <typename C>
set<typename C::value_type> to_set(const C &v) {
	set<typename C::value_type> res;
	for (auto &i : v)
		res.emplace(i);
	return res;
}

#include <iomanip>
template<typename T>
string quote(const T& str) {
	stringstream s;
	s << quoted(str);
	return s.str();
}

#endif //ZAD2_UTIL_H
